/***************************************************************************
 * 
 * Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
 * rights reserved.
 * 
 * Developed  by the Perceptual Computing Section 
 * at the Media Laboratory, MIT, Cambridge, Massachusetts.
 * 
 * Permission to use, copy,  or modify this software and  its documentation
 * for  educational  and  research purposes only and without fee  is hereby
 * granted, provided  that this copyright notice and the original authors's
 * names appear  on all copies and supporting documentation.  If individual
 * files are  separated from  this  distribution directory  structure, this
 * copyright notice must be included.  For any other uses of this software,
 * in original or  modified form, including but not limited to distribution
 * in whole or in  part, specific  prior permission  must be  obtained from
 * MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
 *   basis  of  a commercial  software  or  hardware product  without first
 * obtaining appropriate licenses  from MIT.  MIT. makes no representations
 * about the suitability of this  software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 * 
 **************************************************************************/

// pfGesture.c++

#include "pfGesture.h++"

float plMmtNoise = 1.;
float plAccelNoise = 1.;
int plGrain = 3; // steps per prediction

float pfGesture::bodyMoveXThreshold = 2.;
float pfGesture::postureCrouchThreshold = .9;
float pfGesture::postureSitThreshold = .8;
float pfGesture::handXSmall = .20;
float pfGesture::handXLarge = .38;
float pfGesture::handYSmall = -.15;
float pfGesture::handYLarge = .00;

static char* handGstrings[5] = {
    "Neutral", "Point Down", "Point Side", "Over Head", "Throwing" };

static char* posGstrings[3] = { "Standing", "Bent Over", "Sitting" };

static char *bodyGstrings[2] = { "Stationary", "Moving" };

pfGesture::pfGesture()
{
    Clear();
}

void
pfGesture::Clear()
{
    leftHandAction = handNeutral;
    rightHandAction = handNeutral;
    posture = posStanding;
    bodyAction = still;
    
    vector9Zero( headSave );
    vector9Zero( lhandSave );
    vector9Zero( rhandSave );
    vector9Zero( centSave );
}

static
inline
void
vector3Subtract( float* a, float* b, float* out )
{
    out[0] = a[0] - b[0];
    out[1] = a[1] - b[1];
    out[2] = a[2] - b[2];
}

void
pfGesture::Update( float* head,
		   float* lhand,
		   float* rhand,
		   float* cent,
		   float height )
// each input is a 9-vector, pos(3), vel(3), acc(3)
{
    semaphore.localLOCK();

    // (1) BODY ACTION
    //     -----------
    bodyAction = ( fabsf( cent[0]-centSave[0] ) > bodyMoveXThreshold ) ? moving : still;

    // (2) POSTURE
    //     -------
    float heightFraction = head[1] / height;
    posture = ( heightFraction < postureSitThreshold ) ? posSitting
       : ( heightFraction < postureCrouchThreshold ) ? posCrouched
       : posStanding;

    // (3) HANDS
    //     -----
    float xtend[3];

    vector3Subtract( lhand, head, xtend );
    xtend[0] = fabs( xtend[0] );
    for( int i = 0; i < 3; i++ )
       xtend[i] /= height;
    leftHandAction = ( xtend[0] > handXSmall && xtend[1] < handYSmall ) ? handPointDown
       : ( xtend[0] > handXLarge && xtend[1] < handYLarge ) ? handPointSide
       : ( xtend[1] > handYLarge ) ? handOverHead
       : handNeutral;

    vector3Subtract( rhand, head, xtend );
    xtend[0] = fabs( xtend[0] );
    for( i = 0; i < 3; i++ )
       xtend[i] /= height;
    rightHandAction = ( xtend[0] > handXSmall && xtend[1] < handYSmall ) ? handPointDown
       : ( xtend[0] > handXLarge && xtend[1] < handYLarge ) ? handPointSide
       : ( xtend[1] > handYLarge ) ? handOverHead
       : handNeutral;

    semaphore.localUNLOCK();

    vector9Copy( head, headSave );
    vector9Copy( lhand, lhandSave );
    vector9Copy( rhand, rhandSave );
    vector9Copy( cent, centSave );
}

ostream&
pfGesture::Print( ostream& os )
{
    os << "    BODY: " << bodyGstrings[ bodyAction ] << endl;
    os << " POSTURE: " << posGstrings[ posture ] << endl;
    os << "  L HAND: " << handGstrings[ leftHandAction ] << endl;
    os << "  R HAND: " << handGstrings[ rightHandAction ] << endl;
    os << endl;
    return os;
}
